#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/timex.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/ktime.h>
#include <linux/gpio.h>
#include <linux/device.h>
#include <asm/uaccess.h>	/* for put_user */

#include "capsense.h"

static int major;

int init_module(void);
void cleanup_module(void);
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
        	
#define PIN_SEND 130
#define PIN_RECEIVE_A 183
#define PIN_RECEIVE_B 168

#define DEVICE_NAME "capsense"	

static struct timespec start;
static struct timespec end_a;
static struct timespec end_b;

DECLARE_COMPLETION(completer);
static int got_a, got_b;

static struct class *capsense_class;
static struct device *capsense_dev;

static int device_ioctl(struct inode *inode, struct file *file, 
				unsigned int cmd, unsigned long arg)
{
	int value[2];

	switch (cmd) {
	case CAPSENSE_READ:
	{
#if 0
		struct timeval start;
		struct timeval end;

		/* make sure the capacitor is discharged */
		gpio_set_value(PIN_SEND, 0);
		udelay(200);

		do_gettimeofday(&start);
		gpio_set_value(PIN_SEND, 1);

		while (1) {
			if (gpio_get_value(PIN_RECEIVE))
				break;		
		}

		do_gettimeofday(&end);

		gpio_set_value(PIN_SEND, 0);

		value = (end.tv_sec - start.tv_sec) * 1000000;
		value += (end.tv_usec - start.tv_usec);
#else		
		
		gpio_set_value(PIN_SEND, 0);
		/* make sure the capacitor is discharged, do not call too often */
		udelay(400);

		ktime_get_ts(&start);

		INIT_COMPLETION(completer);

		/* send the wood */
		got_a = got_b = 0;
		
		gpio_set_value(PIN_SEND, 1);

		/* keep a small timeout here as it might happen that we get both interrupts
		 * before we come here. Should we protect the completer with a spinlock or such ? */
		wait_for_completion_interruptible_timeout(&completer, 5);

		if (!got_a || !got_b) {
			printk("capsense timeout !\n");
			return -EPIPE;
		}
		gpio_set_value(PIN_SEND, 0);

		value[0] = (end_a.tv_sec - start.tv_sec) * 1000000000;
		value[0] += (end_a.tv_nsec - start.tv_nsec);
		value[1] = (end_b.tv_sec - start.tv_sec) * 1000000000;
		value[1] += (end_b.tv_nsec - start.tv_nsec);
		
		
#endif
		return copy_to_user((int*)arg, &value[0], sizeof(value));
	}
	}
	return -EINVAL;
}

static struct file_operations fops = {
	.open = device_open,
	.release = device_release,
	.ioctl = device_ioctl,
};

static irqreturn_t _handle_irq_a(int irq, void *ignored)
{	
	ktime_get_ts(&end_a);
	got_a = 1;
	if (got_a && got_b)
		complete(&completer);
	
	return IRQ_HANDLED;
}

static irqreturn_t _handle_irq_b(int irq, void *ignored)
{	
	ktime_get_ts(&end_b);
	got_b = 1;
	if (got_a && got_b)
		complete(&completer);

	return IRQ_HANDLED;
}

/*
 * This function is called when the module is loaded
 */
int init_module(void)
{
        int err = 0;
        
        major = register_chrdev(0, DEVICE_NAME, &fops);

	if (major < 0) {
	  printk(KERN_ALERT "Registering char device failed with %d\n", major);
	  return major;
	}

	if (gpio_request(PIN_SEND, "send") != 0) {
		printk("cannot request send pin\n");
	}
	gpio_direction_output(PIN_SEND, 0);

	if (gpio_request(PIN_RECEIVE_A, "receive_a") != 0) {
		printk("cannot request receive pin A\n");	
	}
	
	err = request_irq(gpio_to_irq(PIN_RECEIVE_A), _handle_irq_a, IRQF_SHARED|IRQF_TRIGGER_RISING, "capsense_a", &end_a) != 0;
	if (err != 0) {
		printk("cannot request irq %d\n", err);	
		goto error_unregister_chrdev;
	}

	if (gpio_request(PIN_RECEIVE_B, "receive_b") != 0) {
		printk("cannot request receive pin B\n");	
	}
	
	err = request_irq(gpio_to_irq(PIN_RECEIVE_B), _handle_irq_b, IRQF_SHARED|IRQF_TRIGGER_RISING, "capsense_b", &end_b) != 0;
	if (err != 0) {
		printk("cannot request irq %d\n", err);	
		goto error_free_irq_a;
	}

	/* creates the devices so that udev makes the correct node */
	capsense_class = class_create(THIS_MODULE, DEVICE_NAME);
        capsense_dev = device_create(capsense_class, NULL, MKDEV(major,0), NULL, DEVICE_NAME);
        
        printk(KERN_INFO "%s major: %d\n", DEVICE_NAME, major);

	return 0;
error_free_irq_a:
	free_irq(gpio_to_irq(PIN_RECEIVE_A), &end_a);
error_unregister_chrdev:
	unregister_chrdev(major, DEVICE_NAME);
	gpio_free(PIN_RECEIVE_A);
	gpio_free(PIN_SEND);

	return -err;
}

void cleanup_module(void)
{	
	free_irq(gpio_to_irq(PIN_RECEIVE_B), &end_b);
	free_irq(gpio_to_irq(PIN_RECEIVE_A), &end_a);

	gpio_free(PIN_RECEIVE_B);
	gpio_free(PIN_RECEIVE_A);
	gpio_free(PIN_SEND);

	device_destroy(capsense_class,MKDEV(major,0));
        class_unregister(capsense_class);
        class_destroy(capsense_class);
        
       	unregister_chrdev(major, DEVICE_NAME);
	printk("Goodbye\n");
}

static int device_open(struct inode *inode, struct file *file)
{
	//printk("capsense: open\n");
	try_module_get(THIS_MODULE);

	return 0;
}

static int device_release(struct inode *inode, struct file *file)
{
	//printk("capsense: close\n");
	module_put(THIS_MODULE);

	return 0;
}

MODULE_LICENSE("GPL"); 
