#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/timex.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/ktime.h>
#include <linux/gpio.h>
#include <linux/device.h>
#include <asm/uaccess.h>	/* for put_user */

#include "max7219.h"

static int major;

static spinlock_t my_lock = SPIN_LOCK_UNLOCKED;

#define MAX_REGISTER_DECODE 0x9
#define MAX_REGISTER_INTENSITY 0xA
#define MAX_REGISTER_SCAN_LIMIT 0xB
#define MAX_REGISTER_SHUTDOWN 0xC
#define MAX_REGISTER_TEST 0xF

#define PIN_LOAD 139
#define PIN_CLOCK 138
#define PIN_DIN 137

#define DBG if(0)

/* schizophrenic logical level due to inverters outside the beagleboard*/
#define HIGH 0
#define LOW 1

static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);

#define DEVICE_NAME "max7219"

static int open;
static int current_digit[4];
static struct class *max7219_class;
static struct device *max7219_dev;

#define DELAY udelay(5)

void _write_register(int addr, u8 value)
{
	int bits = addr << 8 | value;
	int i;
	unsigned long flags;

	DBG printk("write %02x in %02x (bits = %04x)\n", value, addr, bits);

	spin_lock_irqsave(&my_lock, flags);

	gpio_set_value(PIN_CLOCK, LOW);
	gpio_set_value(PIN_LOAD, LOW);
	DELAY;
	
	for (i = 15; i >= 0; i--) {
		gpio_set_value(PIN_DIN, (bits & (1 << i)) ? HIGH : LOW);
		DELAY;
		gpio_set_value(PIN_CLOCK, HIGH);
		DELAY;
		gpio_set_value(PIN_CLOCK, LOW);
		DELAY;
	}

	gpio_set_value(PIN_LOAD, HIGH);
	DELAY;
	gpio_set_value(PIN_LOAD, LOW);
	DELAY;
	
	spin_unlock_irqrestore(&my_lock, flags);	
}

static int device_ioctl(struct inode *inode, struct file *file, 
				unsigned int cmd, unsigned long arg)
{
	Max7219Command command;

	switch (cmd) {
	case MAX7219_COMMAND:
		{
		int addr;
		
		if (copy_from_user(&command, (void __user *)arg, sizeof(command)))
			return -EFAULT;
							
		switch (command.index) {
			case 0: addr = 2; break;
			case 1: addr = 4; break;
			case 2: addr = 3; break;
			case 3: addr = 1; break;
			default: return -EINVAL;		
		}

		DBG printk("%d. 0x%02x\n", addr, command.value);

		if (command.value != current_digit[addr]) {
			_write_register( addr, command.value);
			current_digit[addr] = command.value;
		}
		return 0;
		}
	}
	return -EINVAL;
}

static struct file_operations fops = {
	.open = device_open,
	.release = device_release,
	.ioctl = device_ioctl,
};


/*
 * This function is called when the module is loaded
 */
int init_module(void)
{
        major = register_chrdev(0, DEVICE_NAME, &fops);

	if (major < 0) {
	  printk(KERN_ALERT "Registering char device failed with %d\n", major);
	  return major;
	}

	if (gpio_request(PIN_LOAD, "load") != 0) {
		printk("cannot request load pin\n");
	}
	gpio_direction_output(PIN_LOAD, 0);

	if (gpio_request(PIN_CLOCK, "clock") != 0) {
		printk("cannot request clock pin\n");	
	}	
	gpio_direction_output(PIN_CLOCK, 0);

	if (gpio_request(PIN_DIN, "din") != 0) {
		printk("cannot request din pin\n");	
	}	
	gpio_direction_output(PIN_DIN, 0);

	/* creates the devices so that udev makes the correct node */
	max7219_class = class_create(THIS_MODULE, DEVICE_NAME);
        max7219_dev = device_create(max7219_class, NULL, MKDEV(major,0), NULL, DEVICE_NAME);

	printk(KERN_INFO "%s major: %d device: %p\n", DEVICE_NAME, major, max7219_dev);

	return 0;
}

void cleanup_module(void)
{
	gpio_free(PIN_CLOCK);
	gpio_free(PIN_LOAD);
	gpio_free(PIN_DIN);

	device_destroy(max7219_class,MKDEV(major,0));
        class_unregister(max7219_class);
        class_destroy(max7219_class);

	unregister_chrdev(major, DEVICE_NAME);
	printk("Goodbye\n");
}

static int device_open(struct inode *inode, struct file *file)
{
	int i;
	
	printk("max7219: open\n");
	if (open)
		return -EBUSY;

	try_module_get(THIS_MODULE);

	_write_register(MAX_REGISTER_SHUTDOWN, 1);
	for (i = 1; i <= 4; i++) {
		_write_register(i, 0);
	}
	_write_register(MAX_REGISTER_SCAN_LIMIT, 3);
	_write_register(MAX_REGISTER_TEST, 0);
	_write_register(MAX_REGISTER_DECODE, 0);
	_write_register(MAX_REGISTER_INTENSITY, 0xf);

	open = 1;
	return 0;
}

static int device_release(struct inode *inode, struct file *file)
{
	if (!open)
		return -EINVAL;
	printk("max7219: close\n");

	module_put(THIS_MODULE);

	_write_register(MAX_REGISTER_SHUTDOWN, 0);

	open = 0;
	
	return 0;
}

MODULE_LICENSE("GPL"); 
