#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <alsa/asoundlib.h>
#include <math.h>
#include <inttypes.h>
#include "audio.h"
#include "gpio.h"
#include "timer.h"
#include "message.h"

#define LOG_CONDITION debug_audio
#include "log.h"

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

int debug_audio = 1;

static pthread_t thread_handle;

static snd_pcm_t *playback_handle;

static const char *audio_path;
static const char *audio_string;

static int last_rate;

static int audio_fd = -1;

static unsigned int espeak_sr;

#define SAMPLE_COUNT 4096

int16_t zero[SAMPLE_COUNT];

#define GPIO_MUTE 132

static int mute = 1;
static int mute_time;
static int stop;

static void _configure(unsigned int rate);

unsigned int espeak_init(void);
void espeak_synth(const char * string, int (*output)(int16_t*, int));

static void _mute(void)
{
	if (!mute) {
		log("mute\n");
		gpio_write (GPIO_MUTE, 1);
		mute = 1;
	}
}

static void _unmute(void)
{
	if (mute) {
		log("unmute\n");
		gpio_write (GPIO_MUTE, 0);
		mute = 0;
	}
}

static void _stop()
{
	if (audio_fd != -1)
		close (audio_fd);
	audio_fd = -1;
	
	stop = 0;
	
	mute_time = ze_time() + 200;
}

static int _output_data(int16_t *samples, int count)
{
	int err;
	int retries = 3;
	
try_again:
	if (audio_path || audio_string || stop)
		return 1;

	if ((err = snd_pcm_writei (playback_handle, samples, count)) != count) {
		log_err ("write to audio interface failed %d (%s)\n",
			 err, snd_strerror (err));

                err = snd_pcm_recover(playback_handle, err, 0);
		if (err < 0) {
			log_err ("cannot recover %d (%s)\n",
				 err, snd_strerror (err));
		}
		if (retries-- > 0)
			goto try_again;
	}
	
	return 0;		
}

static void *audio_thread(void*data)
{
	int16_t samples[SAMPLE_COUNT];
	int len;
	int was_playing = 0;
	
	while (1) {
		if (mute_time && ze_time() > mute_time) {
			_mute();
			mute_time = 0;
		}
				
		if (audio_path) {
			_unmute();
			mute_time = 0;

			was_playing = 1;
			_configure(44100);
			
			audio_fd = open(audio_path, O_RDONLY);
			if (audio_fd == -1) {
				printf("cannot open %s\n", audio_path);
				audio_path = NULL;
				continue;
			}

			audio_path = NULL;

			while ((len = read (audio_fd, samples, sizeof(samples))) > 0)
			{
				if (_output_data(samples, len / 2)) {
					log ("abort\n");
					break;
				}
			}
			
			_stop();
			continue;
			
		}
		
		if (audio_string) {
			char *string;
			
			was_playing = 1;
			
			pthread_mutex_lock(&mutex);
			string = (char*)audio_string;
			audio_string = NULL;
			pthread_mutex_unlock(&mutex);
			
			_unmute();
			mute_time = 0;

			_configure(espeak_sr);

			espeak_synth(string, _output_data);

			free(string);
			
			_stop();
			continue;
		}
		
		if (was_playing) {
			message_put(MESSAGE_SOUND_STOPPED);
			was_playing = 0;
		}
		//_output_data(zero, SAMPLE_COUNT);
		
		usleep (1 * 000);
	}
}

static void _configure(unsigned int rate)
{
	int err;
	snd_pcm_hw_params_t *hw_params;
	snd_pcm_uframes_t buffer_size = SAMPLE_COUNT;		
	   
	log("configure %d\n", rate);
	
	if (rate == last_rate)
		return;

	last_rate = rate;
	
	if (playback_handle) {
		log("close previous device\n");
		snd_pcm_drain(playback_handle);
		snd_pcm_pause(playback_handle, 1);
		if (snd_pcm_close (playback_handle)) {
			log("cannot close audio device\n");
		}
		playback_handle = NULL;
	}
	
try_again:
	if ((err = snd_pcm_open (&playback_handle, "default", SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
		fprintf (stderr, "cannot open audio device default (%s)\n", 
			 snd_strerror (err));
		usleep(50 * 1000);
		goto try_again;
	}

	if ((err = snd_pcm_hw_params_malloc (&hw_params)) < 0) {
		fprintf (stderr, "cannot allocate hardware parameter structure (%s)\n",
			 snd_strerror (err));
	}
			 
	if ((err = snd_pcm_hw_params_any (playback_handle, hw_params)) < 0) {
		fprintf (stderr, "cannot initialize hardware parameter structure (%s)\n",
			 snd_strerror (err));
	}

	if ((err = snd_pcm_hw_params_set_access (playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
		fprintf (stderr, "cannot set access type (%s)\n",
			 snd_strerror (err));
	}

	if ((err = snd_pcm_hw_params_set_format (playback_handle, hw_params, SND_PCM_FORMAT_S16_LE)) < 0) {
		fprintf (stderr, "cannot set sample format (%s)\n",
			 snd_strerror (err));
	}

	if ((err = snd_pcm_hw_params_set_rate_near (playback_handle, hw_params, &rate, 0)) < 0) {
		fprintf (stderr, "cannot set sample rate (%s)\n",
			 snd_strerror (err));
	}

	if ((err = snd_pcm_hw_params_set_channels (playback_handle, hw_params, 1)) < 0) {
		fprintf (stderr, "cannot set channel count (%s)\n",
			 snd_strerror (err));
	}

	err = snd_pcm_hw_params_set_buffer_size_near(playback_handle, hw_params, &buffer_size);
	if (err < 0) {
		printf("Unable to set buffer size %d for playback: %s\n", (int)buffer_size, snd_strerror(err));
	}

	if ((err = snd_pcm_hw_params (playback_handle, hw_params)) < 0) {
		fprintf (stderr, "cannot set parameters (%s)\n",
			 snd_strerror (err));
	}
	
	snd_pcm_hw_params_free (hw_params);

	if ((err = snd_pcm_prepare (playback_handle)) < 0) {
		fprintf (stderr, "cannot prepare audio interface for use (%s)\n",
			 snd_strerror (err));
	}
}

void audio_init(void)
{
	gpio_setup(GPIO_MUTE, GPIO_OUTPUT, 0);

	_configure(44100);
	
	espeak_sr = espeak_init();
	
	log ("espeake sample rate : %d\n", espeak_sr);
	
	pthread_create(&thread_handle, NULL, audio_thread, NULL);

	audio_play_string("Bonjour Noémie ! Comment ça va ?");
}

void audio_play_file(const char *file_path)
{
	log("audio_play_file %s\n", file_path);
	audio_path = file_path;
}

void audio_play_string(const char *string)
{
	log("audio_play_string %s\n", string);

	pthread_mutex_lock(&mutex);
	if (audio_string) {
		free ((void*)audio_string);
		audio_string = NULL;
	}
	pthread_mutex_unlock(&mutex);
	
	audio_string = strdup(string);
}

void audio_stop(void)
{
	log("audio_stop\n");
	stop = 1;
}

