#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdint.h>

#include "gpio.h"
#define LOG_CONDITION debug_gpio
#include "log.h"

int debug_gpio;

static int pad_fd;
static int gpio_fd;
static uint32_t *padconf;
static uint32_t *gpioconf;

int gpio_init(void) 
{
	log("gpio init...");

	pad_fd = open("/dev/mem", O_RDWR | O_SYNC);

	if (pad_fd < 0) {
		printf("cannot open pad_fd\n");
		return -1;
	}

	padconf = mmap(NULL, 0x10000, PROT_READ | PROT_WRITE, MAP_SHARED, pad_fd, 0x48000000);
	if (padconf == MAP_FAILED) {
		printf("cannot map padconf\n");
		goto exit_close_pad_fd;
	}

	gpio_fd = open("/dev/mem", O_RDWR | O_SYNC);
  	if (gpio_fd == -1) {
		printf("cannot open gpio_fd\n");
		goto exit_unmap_padconf;
	}

	/* we do not map bank 1 */
	gpioconf = mmap(NULL, 0x10000, PROT_READ | PROT_WRITE, MAP_SHARED, gpio_fd, 0x49050000);
	if (gpioconf == MAP_FAILED) {
		printf("cannot map gpioconf\n");
		goto exit_close_gpio_fd;
	}

	log("ok\n");
	return 0;

exit_close_gpio_fd:
	close(gpio_fd);
exit_unmap_padconf:
	munmap(padconf, 0x10000);
exit_close_pad_fd:
	close(pad_fd);

	return -1;
}

static void _get_bank(int gpio, uint32_t **base, int *bit)
{
	int bank = gpio / 32;

	if (bank < 1) {
		printf("bad gpio bank\n");	
	}

	*base = &gpioconf[((bank - 1) * 0x2000) / 4];
	*bit = gpio % 32;
}

void gpio_setup(int gpio, int is_out, int pull_up)
{
	uint32_t *pad;
	uint32_t *oe;
	uint32_t conf_value = 0;
	uint32_t old_value = 0;
	int use_high_bits = gpio & 1;
	int bit;

	log("setup GPIO %3d as %s\n", gpio, is_out ? "OUTPUT": "INPUT");

	if (gpio < 128) {
		printf("not handled yet\n");
		return;	
	} else if (gpio < 167) {
		pad = &padconf[0x2154 / 4 + (gpio - 128) / 2];
	} else if (gpio == 168) {
		pad = &padconf[0x21BC / 4];
		use_high_bits = 1;
	} else if (gpio == 183) {
		pad = &padconf[0x21C0 / 4];
		use_high_bits = 0;			
	} else {
		printf("not handled yet\n");
		return;	
	}

	_get_bank(gpio, &oe, &bit);

	oe += 0x34 / 4;
	
	old_value = *pad;

	/* PAD config */
	conf_value |= 4;
	if (!is_out) {
		conf_value |= 1 << 8; /* input enable */
		if (pull_up == GPIO_PULLUP) {
			conf_value |= 1 << 4; /* use pull up */
		}
			
		if (pull_up != GPIO_NONE) {
			conf_value |= 1 << 3; /* enable the pull up/down */	
		}
	}

	if (use_high_bits) {		
		old_value &= 0xffff;
		*pad = old_value | (conf_value << 16);
	} else {
		old_value &= 0xffff << 16;
		*pad = old_value | conf_value;
	}

	/* GPIO config */
	if (is_out) {
		*oe &= ~(1 << bit);
	} else {
		*oe |= (1 << bit);
	}
}

void gpio_write(int gpio, int value)
{
	uint32_t *dataout;
	int bit;

	_get_bank(gpio, &dataout, &bit);

	dataout += 0x3C / 4;

	log("writing GPIO %3d <= %d\n", gpio, value);

	if (value) 
		*dataout |= 1 << bit;
	else
		*dataout &= ~(1 << bit);
}

int gpio_read(int gpio)
{
	uint32_t volatile*datain;
	int bit;
	int value;

	_get_bank(gpio, (uint32_t**)&datain, &bit);

	datain += 0x38 / 4;

	value = (*datain & (1 << bit));

	log("reading GPIO %3d => %d\n", gpio, !!value);
	return !!value;
}

void gpio_exit(void)
{
	log("gpio_exit\n");

	munmap(gpioconf, 0x10000);
	munmap(padconf, 0x10000);
	close(pad_fd);
	close(gpio_fd);
}
