#include <stdint.h>
#include <inttypes.h>
#include <stdio.h>
#include <time.h>
#include <pthread.h>
#include "timer.h"

#define MAX_TIMER 32

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

typedef struct {
	int id;
	void (*cb)(void *priv);
	void *priv;
	int64_t timeout;
	int interval;
	TimerMode mode;
} Timer;

static Timer timer[MAX_TIMER];
static int id_counter = 1;

void timer_init(void)
{
	int i;
	for (i = 0; i < MAX_TIMER; i++) {
		timer[i].id = -1;
	}
}

volatile int64_t ze_time(void)
{
	static struct timespec tv_first;
	static int initialized;
	struct timespec tv;

	clock_gettime(CLOCK_MONOTONIC, &tv );

	if (!initialized) {
		tv_first = tv;
		initialized = 1;
	}
	
	return (tv.tv_sec - tv_first.tv_sec) * 1000 + (tv.tv_nsec - tv_first.tv_nsec) / 1000000;
}

int timer_add (void (*cb)(void *priv), void *priv, TimerMode mode, int delay, int interval)
{
	int i;
	
	pthread_mutex_lock(&mutex);
	
	for (i = 0; i < MAX_TIMER; i++) {
		if (timer[i].id == -1) {
			int id;
			
			timer[i].id = id_counter++;
			timer[i].cb = cb;
			timer[i].priv = priv;
			timer[i].mode = mode;
			timer[i].timeout = ze_time() + delay;
			timer[i].interval = interval;
			//printf ("found timer at %d\n", i);
			id = timer[i].id;
			
			pthread_mutex_unlock(&mutex);

			return id;
		}
	}
	
	pthread_mutex_unlock(&mutex);
	return -1;
}

void timer_remove (int *id)
{
	int i;

	pthread_mutex_lock(&mutex);

	if (*id == -1)
		goto end;
	
	for (i = 0; i < MAX_TIMER; i++) {
		if (timer[i].id == *id) {
			timer[i].id = -1;
			break;
		}
	}
	*id = -1;
end:
	pthread_mutex_unlock(&mutex);
}

void timer_trigger (void)
{
	int i;
	
	pthread_mutex_lock(&mutex);

	for (i = 0; i < MAX_TIMER; i++) {
		if (timer[i].id != -1 && timer[i].timeout < ze_time()) {
			void (*cb)(void*);
			void *priv;
			
			if (timer[i].mode == TimerSingle) {
				timer[i].id = -1;
			} else {
				timer[i].timeout = ze_time() + timer[i].interval;
			}

			cb = timer[i].cb;
			priv = timer[i].priv;
			
			pthread_mutex_unlock(&mutex);
			if (cb)
				cb(priv);
			pthread_mutex_lock(&mutex);
		}
	}	
	pthread_mutex_unlock(&mutex);
}

int64_t timer_next_timeout(void)
{
	int64_t min = INT64_MAX;
	int i;
	
	pthread_mutex_lock(&mutex);
	for (i = 0; i < MAX_TIMER; i++) {
		if (timer[i].id != -1 && timer[i].timeout < min) {
			min = timer[i].timeout;
		}
	}	
	pthread_mutex_unlock(&mutex);

	return min;
}
